/*
 * Copyright (c) 1998 Kungliga Tekniska Hgskolan
 * (Royal Institute of Technology, Stockholm, Sweden).
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Kungliga Tekniska
 *      Hgskolan and its contributors.
 * 
 * 4. Neither the name of the Institute nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
RCSID("$Id: tl.c,v 1.2 1998/02/12 21:10:00 lha Exp $");
#endif

#include "tl_locl.h"

static char *tl_errorstrings[] = {
    "Success",
    "Unknown command",
    "Memory/allocation error",
    "Invalid number of arguments",
    "Can't parse"
};

static TL_cmd *
tl_match (TL_cmd *cmds, char *cmd, int exactp)
{
    TL_cmd *c, *current = NULL, *partial_cmd = NULL;
    int partial_match = 0;

    for (c = cmds; c->name; ++c) {
	if (c->usage)
	    current = c;
	if (strcmp (cmd, c->name) == 0)
	    return current;
	else if (strncmp (cmd, c->name, strlen(cmd)) == 0 &&
		 partial_cmd != current) {
	    ++partial_match;
	    partial_cmd = current;
	}
    }
    if (partial_match == 1 && !exactp)
	return partial_cmd;
    else
	return NULL;
}

static TL_value*
tl_alloc(char *name, char **value, int num)
{
    TL_value *val;
    int i;
    int err;

    if (num < 0) 
	return NULL ;

    val =  malloc(sizeof(TL_value));
    
    if (val == NULL)
	return NULL;

    val->value = malloc(sizeof(char*) * (num + 1));

    val->name = strdup(name);
    i = 0 ;
    while(num > i && !err) {
	val->value[i] = strdup(value[i]);
	if (val->value[i] == NULL)
	    err = 1;
	i++;
    }
    val->value[i] = NULL;

    val->next = NULL;

    if (val->name == NULL || err) {
	if (val->name)
	    free(val->name);
	while(i >= 0)
	    free(val->value[i--]);
	free(val);
	val = NULL;
    }

    return val;
}

static TL_cmd*
tl_findnum(TL_cmd *cmds, int num)
{
    int cnt = 0 ;

    if (num < 0)
	return NULL; 

    while (cmds) {
	if (cmds->usage)
	    ++cnt;
	if (num+1 == cnt)
	    return cmds;
	++cmds;
    }
    return NULL;
}


void 
tl_help (TL_cmd *cmds, char *cmdname, int verbose)
{
    TL_cmd *c, *prev_c;
    
    printf("%s ", cmdname);
    prev_c = NULL;
    c = cmds;
    while(c->name) {
	if (c->usage) 
	    printf ("%s-%s %s %s",verbose?"\n\t":"", c->name,c->usage, 
		    verbose && c->verbosehelp?c->verbosehelp:"");
	++c;
    }
    printf("\n");
}

void
tl_free (TL_value *val)
{
    TL_value *tmp;
    while(val) {
	tmp = val->next;
	free(val->name);
	while(*val->value) {
	    free(*val->value);
	    val->value ++;
	}
	free(val->value) ;
	free(val);
	val = tmp;
    }
}

static char *tl_magic_cookie[] = { "magic-cookie", NULL };

int
tl_loop (TL_cmd *cmds, int argc, char **argv, TL_value **tl_ret)
{
    int cmdcnt = 1, err = 0, dashless = 1, dashcnt = 0;
    TL_cmd *cmd;
    TL_value *cur;
    char *tmp;

    *tl_ret = tl_alloc("4711 != 17", tl_magic_cookie, 1);
    if (*tl_ret == NULL)
	return TL_MEMERR;
    
    cur = *tl_ret;

    while (cmdcnt < argc) {
	if (*argv[cmdcnt] != '-') {
	    if (!dashless) {
		err = TL_CNPARSE;
		break;
	    }
	    cmd = tl_findnum(cmds, dashcnt);
	    ++dashcnt;
	    --cmdcnt;

	} else {
	    dashless = 0;

	    /* Allow - as an argument */
	    if (argv[cmdcnt][0] == '-' && argv[cmdcnt][1] == '\0')
		tmp = argv[cmdcnt];
	    else
		tmp = &argv[cmdcnt][1];
	    cmd = tl_match(cmds, tmp, 0);
	}

	if (cmd == NULL) {
	    err = TL_UNKNOWNCMD;
	    break;
	}
	
	if (argc <= cmdcnt + cmd->numargs) {
	    err = TL_NUMARGS;
	    break ;
	}
	
	cur->next = tl_alloc(cmd->name, &argv[cmdcnt]+1, cmd->numargs);
	
	cur = cur->next;
	if (cur == NULL) {
	    err = TL_MEMERR;
	    break;
	}
	cmdcnt += cmd->numargs + 1;
    }
	
    if (err) {
	tl_free(*tl_ret);
	*tl_ret = NULL;
    }

    return err;
}


void
tl_print(TL_value *val) 
{
    char **tmp;
    if (val) val = val->next;  /* Skip the magic cookie */
    while(val) {
	printf("%s\t", val->name);
	tmp = val->value ;
	while(*tmp) {
	    printf("%s", *tmp);
	    if (*(tmp+1) != NULL)
		printf(", ");
	    tmp++;
	}
	printf("\n");
	val = val->next;
    }
}


char *
tl_errstr(int i)
{
    if (i < 0 || i >= (sizeof(tl_errorstrings)/sizeof(char*)))
	return "Unknown errorcode";

    return tl_errorstrings[i];

}

char **
tl_find(TL_value *val, char *name)
{
    if (name == NULL || val == NULL)
	return NULL;

    while (val) {
	if (strcmp(val->name, name) == 0) {
	    return val->value;
	}
	val = val->next;
    }
    return NULL;
}

